# The Practice module
module Practice
  # The Implementation module
  module Implementation
    # Brainfuck interpreter written in Ruby.
    class Interpreter
      # The Brainfuck command.
      Command = Struct.new(:symbol)

      # Increment the data pointer (to point to the next cell to the right).
      @lshift = Command.new('<')
      def @lshift.process(cxt, _)
        cxt.cursor -= 1
        raise ArgumentError.new('Subscript out of range') if cxt.cursor < 0
      end

      # Decrement the data pointer (to point to the next cell to the left).
      @rshift = Command.new('>')
      def @rshift.process(cxt, _)
        cxt.cursor += 1
        cxt.memory << 0 if cxt.cursor == cxt.memory.length
      end

      # Increment (increase by one) the byte at the data pointer.
      @plus = Command.new('+')
      def @plus.process(cxt, _)
        value = cxt.memory[cxt.cursor] + 1
        cxt.memory[cxt.cursor] = value > 255 ? 0 : value
      end

      # Decrement (decrease by one) the byte at the data pointer.
      @minus = Command.new('-')
      def @minus.process(cxt, _)
        value = cxt.memory[cxt.cursor] - 1
        cxt.memory[cxt.cursor] = value < 0 ? 255 : value
      end

      # Output the byte at the data pointer.
      @dot = Command.new('.')
      def @dot.process(cxt, _)
        value = cxt.memory[cxt.cursor]
        cxt.output(value.chr)
      end

      # Accept one byte of input,
      # storing its value in the byte at the data pointer.
      @comma = Command.new(',')
      def @comma.process(cxt, _)
        value = cxt.input
        cxt.memory[cxt.cursor] = value.ord
      end

      # If the byte at the data pointer is zero,
      # then instead of moving the instruction pointer
      # forward to the next command,
      # jump it forward to the command after the matching ] command.
      @lbracket = Command.new('[')
      def @lbracket.process(cxt, bracemap)
        cxt.counter = bracemap[cxt.counter] if cxt.memory[cxt.cursor] == 0
      end

      # If the byte at the data pointer is nonzero,
      # then instead of moving the instruction pointer
      # forward to the next command,
      # jump it back to the command after the matching [ command.
      @rbracket = Command.new(']')
      def @rbracket.process(cxt, bracemap)
        cxt.counter = bracemap[cxt.counter] if cxt.memory[cxt.cursor] != 0
      end

      @symbols = %i(
        lshift rshift plus minus dot comma lbracket rbracket
      ).reduce({}) do |symbols, name|
        command = instance_variable_get("@#{name}")
        symbols[command.symbol] = command
        symbols
      end

      # Initialize.
      def initialize(source)
        syms = Interpreter.symbols
        tokens = source.each_char.map { |c| c }.select { |c| syms.include?(c) }
        @commands = tokens.map { |c| Interpreter.command(c) }
        @bracemap = parse(@commands)
      end

      # To evaluate the source code.
      def evaluate(cxt)
        while cxt.counter < @commands.length
          command = @commands[cxt.counter]
          command.process(cxt, @bracemap)
          cxt.counter += 1
        end
      end

      private

      # Parse to brainfuck code.
      def parse(commands)
        bracestack, bracemap = [], {}
        commands.each_with_index do |command, index|
          if command.symbol == '['
            bracestack << index
          elsif command.symbol == ']'
            message = "syntax error, unexpected ']'"
            raise ArgumentError.new message if bracestack.empty?
            bracepos = bracestack.pop
            bracemap[bracepos] = index
            bracemap[index] = bracepos
          end
        end

        message = "syntax error, unexpected '['"
        raise ArgumentError.new message unless bracestack.empty?
        bracemap
      end

      class << self
        # Get brainfuck symbols
        def symbols
          @symbols.keys
        end

        def command(c)
          @symbols[c]
        end
      end
    end
  end
end
